package net.sf.itcb.common.client.aop;

import java.io.IOException;
import java.util.List;
import java.util.ArrayList;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import net.sf.itcb.common.business.exceptions.CommonItcbException;
import net.sf.itcb.common.client.exceptions.ClientItcbException;
import net.sf.itcb.common.client.exceptions.ClientItcbExceptionMappingErrors;
import net.sf.itcb.common.client.marshaller.ItcbJaxb2Marshaller;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.ws.client.core.SourceExtractor;
import org.springframework.ws.client.core.WebServiceMessageCallback;
import org.springframework.ws.client.core.support.WebServiceGatewaySupport;
import org.springframework.ws.client.support.interceptor.ClientInterceptor;
import org.springframework.ws.soap.client.SoapFaultClientException;
import org.springframework.xml.transform.StringResult;
import org.springframework.xml.transform.StringSource;

/**
 * This interceptor allows to use generic client in order to call a remote provider webservice.
 * It can manage each interface which respects those rules :
 * <ul>
 * <li>Only one parameter <b>*Request</b> generated by JAXB from an element or from a complexType defined in a WSDL. This object contains all information needed by the service</li>
 * <li>A return object <b>*Response</b> generated by JAXB  from an element or from a complexType defined in a WSDL</li>
 * </ul>
 * If it doesn't match, a {@link net.sf.itcb.common.client.exceptions.ClientItcbException} is thrown.<br/>
 * If requests and responses are built from a complexType, so no {@link XmlRootElement} is defined in JAXB classes. <br/>
 * For this context, you have two choices :<br/>
 * <ul>
 *   <li><b>Choice 1 :</B> to define the {@link QName} that has to be used for requests.</li>
 *   <li><b>Choice 2 :</B> to let this class generate the {@link QName} using the class as localPart.</li>
 * </ul>
 * This {@link QName} is used for main xml tag in SOAP body.<br/>
 * This {@link QName} has to match the declaration in your provider wsdl.
 * For example 
 * <ul>
 * <li> When the wsdl part name doesn't match to the name of the wsdl message, you have to use Option 1 : </li>
 * <PRE>{@code
 * <wsdl:message name="getSomethingRequest">
 *   <wsdl:part name="parameters" type="GetSomethingRequest"></wsdl:part>
 * </wsdl:message>
 * <wsdl:message name="getSomethingResponse">
 *   <wsdl:part name="parameters" type="ns1:GetSomethingResponse"></wsdl:part>
 * </wsdl:message>
 *
 * }</PRE>
 * <li> When the wsdl part name matches to the name of the wsdl message, you should use Option 2 : </li>
 * <PRE>{@code
 * <wsdl:message name="getSomethingRequest">
 *   <wsdl:part name="getSomethingRequest" type="GetSomethingRequest"></wsdl:part>
 * </wsdl:message>
 * <wsdl:message name="getSomethingResponse">
 *   <wsdl:part name="getSomethingRequest" type="ns1:GetSomethingResponse"></wsdl:part>
 * </wsdl:message>
 *
 * }</PRE>
 * </ul>
 * 
 * You have to use {@link ItcbJaxb2Marshaller} in order this class can retrieve the original marshaller and unmarshaller which allow to unmarshall with a specified returned type
 * 
 * @author Pierre Le Roux
 */
@SuppressWarnings({ "unchecked", "rawtypes" })
public class ItcbProviderInterceptor extends WebServiceGatewaySupport implements MethodInterceptor, ApplicationContextAware {

	protected org.springframework.context.MessageSource messageSource;
	
	protected QName xmlRootElementTag;
	
	protected WebServiceMessageCallback webserviceMessageCallback;
	
	protected List<ClientInterceptorTransportExceptionListener> clientInterceptorTransportExceptionListener;
	
	
	@Override
	public void setApplicationContext(ApplicationContext applicationContext)
			throws BeansException {
		messageSource=applicationContext.getBean("common-clientResources", org.springframework.context.MessageSource.class);
	}
			
	public void setXmlRootElementTag(QName xmlRootElementTag) {
		this.xmlRootElementTag=xmlRootElementTag;
	}
	
	@Override
	protected void initGateway() throws Exception {
		clientInterceptorTransportExceptionListener = new ArrayList<ClientInterceptorTransportExceptionListener>();
		if( this.getInterceptors() != null ) {
			for (ClientInterceptor clientInterceptor : this.getInterceptors()) {
				if(clientInterceptor instanceof ClientInterceptorWebserviceTemplateAware) {
					((ClientInterceptorWebserviceTemplateAware)clientInterceptor).setWebserviceTemplate(getWebServiceTemplate());
				}
				if(clientInterceptor instanceof ClientInterceptorTransportExceptionListener) {
					clientInterceptorTransportExceptionListener.add((ClientInterceptorTransportExceptionListener)clientInterceptor);
				}
			}
		}
	}
	
	/**
	 * It is possible to add a callback, for example in order to add a custom soap action in header
	 * Not mandatory
	 */
	public void setWebserviceMessageCallback(WebServiceMessageCallback webserviceMessageCallback) {
		this.webserviceMessageCallback = webserviceMessageCallback;
	}
	
	public Object invoke(MethodInvocation arg0) throws Throwable {
		if(arg0.getArguments().length != 1) {
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_NOT_CORRECT_INTERFACE, messageSource.getMessage("common-client.exception.not_correct_interface", new Object[] {arg0.getMethod()}, LocaleContextHolder.getLocale()));
		}
		
		try  {
			StringResult stringResult= new StringResult();
			Object requestParameters= arg0.getArguments()[0];
			Class<?> requestParametersClass = requestParameters.getClass();
			if(requestParametersClass.isAnnotationPresent(XmlRootElement.class)) {
				((ItcbJaxb2Marshaller)getWebServiceTemplate().getMarshaller()).getMarshaller().marshal(requestParameters, stringResult);
			}
			else {
				if(xmlRootElementTag!= null) {
					((ItcbJaxb2Marshaller)getWebServiceTemplate().getMarshaller()).getMarshaller().marshal(new JAXBElement(xmlRootElementTag, requestParametersClass, requestParameters), stringResult);
				}
				else {
					((ItcbJaxb2Marshaller)getWebServiceTemplate().getMarshaller()).getMarshaller().marshal(new JAXBElement(new QName(requestParametersClass.getSimpleName()), requestParametersClass, requestParameters), stringResult);
				}
			}
			if( webserviceMessageCallback == null) {
					return getWebServiceTemplate().sendSourceAndReceive(new StringSource(stringResult.getWriter().toString()), new ItcbResponseExtractor(arg0.getMethod().getReturnType()));
			}
			else {
				return getWebServiceTemplate().sendSourceAndReceive(new StringSource(stringResult.getWriter().toString()), webserviceMessageCallback, new ItcbResponseExtractor(arg0.getMethod().getReturnType()));
			}
		}
		catch (CommonItcbException cie) { 
			throw cie;
		}
		catch (SoapFaultClientException sfce) { 
			logger.error(sfce.getFaultCode(), sfce);
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_TECHNICAL_ERROR, messageSource.getMessage("common-client.exception.technical_error", new Object[] {arg0.getMethod(), getWebServiceTemplate().getDestinationProvider().getDestination(), sfce.getMessage()}, LocaleContextHolder.getLocale()), sfce);
		}
		catch(java.lang.Throwable e) {
			logger.error(e.getMessage(), e);
			if(getInterceptors() != null) {
				for (ClientInterceptorTransportExceptionListener clientInterceptor : clientInterceptorTransportExceptionListener) {
					clientInterceptor.handleTransportException(e, arg0.getArguments()[0]);
				}
			}
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_TECHNICAL_ERROR, messageSource.getMessage("common-client.exception.technical_error", new Object[] {arg0.getMethod(), getWebServiceTemplate().getDestinationProvider().getDestination(), e.getMessage()}, LocaleContextHolder.getLocale()), e);
		}
		
	}
	
	
	/**
	 * This class is an extractor for the response
	 * It allows to unmarshall the response using a specified returned type which is the expected type for the method intercepted by the MethodInterceptor 
	 * @author Pierre Le Roux
	 * @param <T>
	 */
	public class ItcbResponseExtractor<T> implements SourceExtractor<T>{

		private Class<T> clazz;
		
		public ItcbResponseExtractor(Class<T> clazz) {
			super();
			this.clazz=clazz;
		}
		
		@Override
		public T extractData(Source src) throws IOException,
				TransformerException {
			try {
				Object object = ((ItcbJaxb2Marshaller)getWebServiceTemplate().getUnmarshaller()).getUnmarshaller().unmarshal(src, clazz);
				if(object instanceof JAXBElement) {
					return (T)((JAXBElement)object).getValue();
				}
				else {
					return (T)object;
				}
			} catch (JAXBException e) {
				throw new TransformerException(e);
			}

		}
		
	}	
}
